@@ -2,6 +2,7 @@
 
 # Copyright (c) IPython Development Team.
 # Distributed under the terms of the Modified BSD License.
+from __future__ import annotations
 
 import asyncio
 import concurrent.futures
@@ -11,6 +12,7 @@ import logging
 import os
 import socket
 import sys
+import threading
 import time
 import typing as t
 import uuid
@@ -19,6 +21,8 @@ from datetime import datetime
 from functools import partial
 from signal import SIGINT, SIGTERM, Signals, default_int_handler, signal
 
+from .control import CONTROL_THREAD_NAME
+
 if sys.platform != "win32":
     from signal import SIGKILL
 else:
@@ -57,14 +61,21 @@ from zmq.eventloop.zmqstream import ZMQStream
 from ipykernel.jsonutil import json_clean
 
 from ._version import kernel_protocol_version
+from .iostream import OutStream
 
 
-def _accepts_cell_id(meth):
+def _accepts_parameters(meth, param_names):
     parameters = inspect.signature(meth).parameters
-    cid_param = parameters.get("cell_id")
-    return (cid_param and cid_param.kind == cid_param.KEYWORD_ONLY) or any(
-        p.kind == p.VAR_KEYWORD for p in parameters.values()
-    )
+    accepts = {param: False for param in param_names}
+
+    for param in param_names:
+        param_spec = parameters.get(param)
+        accepts[param] = (
+            param_spec
+            and param_spec.kind in [param_spec.KEYWORD_ONLY, param_spec.POSITIONAL_OR_KEYWORD]
+        ) or any(p.kind == p.VAR_KEYWORD for p in parameters.values())
+
+    return accepts
 
 
 class Kernel(SingletonConfigurable):
@@ -77,7 +88,7 @@ class Kernel(SingletonConfigurable):
     # attribute to override with a GUI
     eventloop = Any(None)
 
-    processes: t.Dict[str, psutil.Process] = {}
+    processes: dict[str, psutil.Process] = {}
 
     @observe("eventloop")
     def _update_eventloop(self, change):
@@ -90,7 +101,7 @@ class Kernel(SingletonConfigurable):
     profile_dir = Instance("IPython.core.profiledir.ProfileDir", allow_none=True)
     shell_stream = Instance(ZMQStream, allow_none=True)
 
-    shell_streams = List(
+    shell_streams: List[t.Any] = List(
         help="""Deprecated shell_streams alias. Use shell_stream
 
         .. versionchanged:: 6.0
@@ -111,8 +122,7 @@ class Kernel(SingletonConfigurable):
         )
         if self.shell_stream is not None:
             return [self.shell_stream]
-        else:
-            return []
+        return []
 
     @observe("shell_streams")
     def _shell_streams_changed(self, change):  # pragma: no cover
@@ -150,10 +160,10 @@ class Kernel(SingletonConfigurable):
 
     # This should be overridden by wrapper kernels that implement any real
     # language.
-    language_info: t.Dict[str, object] = {}
+    language_info: dict[str, object] = {}
 
     # any links that should go in the help menu
-    help_links = List()
+    help_links: List[dict[str, str]] = List()
 
     # Experimental option to break in non-user code.
     # The ipykernel source is in the call stack, so the user
@@ -177,7 +187,7 @@ class Kernel(SingletonConfigurable):
 
     # track associations with current request
     _allow_stdin = Bool(False)
-    _parents = Dict({"shell": {}, "control": {}})
+    _parents: Dict[str, t.Any] = Dict({"shell": {}, "control": {}})
     _parent_ident = Dict({"shell": b"", "control": b""})
 
     @property
@@ -187,7 +197,7 @@ class Kernel(SingletonConfigurable):
             DeprecationWarning,
             stacklevel=2,
         )
-        return self.get_parent(channel="shell")
+        return self.get_parent()
 
     # Time to sleep after flushing the stdout/err buffers in each execute
     # cycle.  While this introduces a hard limit on the minimal latency of the
@@ -263,6 +273,13 @@ class Kernel(SingletonConfigurable):
     def __init__(self, **kwargs):
         """Initialize the kernel."""
         super().__init__(**kwargs)
+
+        # Kernel application may swap stdout and stderr to OutStream,
+        # which is the case in `IPKernelApp.init_io`, hence `sys.stdout`
+        # can already by different from TextIO at initialization time.
+        self._stdout: OutStream | t.TextIO = sys.stdout
+        self._stderr: OutStream | t.TextIO = sys.stderr
+
         # Build dict of handlers for message types
         self.shell_handlers = {}
         for msg_type in self.msg_types:
@@ -274,6 +291,11 @@ class Kernel(SingletonConfigurable):
 
         self.control_queue: Queue[t.Any] = Queue()
 
+        # Storing the accepted parameters for do_execute, used in execute_request
+        self._do_exec_accepted_params = _accepts_parameters(
+            self.do_execute, ["cell_meta", "cell_id"]
+        )
+
     def dispatch_control(self, msg):
         self.control_queue.put_nowait(msg)
 
@@ -288,7 +310,7 @@ class Kernel(SingletonConfigurable):
 
     async def _flush_control_queue(self):
         """Flush the control queue, wait for processing of any pending messages"""
-        tracer_future: t.Union[concurrent.futures.Future[object], asyncio.Future[object]]
+        tracer_future: concurrent.futures.Future[object] | asyncio.Future[object]
         if self.control_thread:
             control_loop = self.control_thread.io_loop
             # concurrent.futures.Futures are threadsafe
@@ -301,7 +323,8 @@ class Kernel(SingletonConfigurable):
 
         def _flush():
             # control_stream.flush puts messages on the queue
-            self.control_stream.flush()
+            if self.control_stream:
+                self.control_stream.flush()
             # put Future on the queue after all of those,
             # so we can wait for all queued messages to be processed
             self.control_queue.put(tracer_future)
@@ -311,11 +334,13 @@ class Kernel(SingletonConfigurable):
 
     async def process_control(self, msg):
         """dispatch control requests"""
+        if not self.session:
+            return
         idents, msg = self.session.feed_identities(msg, copy=False)
         try:
             msg = self.session.deserialize(msg, content=True, copy=False)
         except Exception:
-            self.log.error("Invalid Control Message", exc_info=True)
+            self.log.error("Invalid Control Message", exc_info=True)  # noqa: G201
             return
 
         self.log.debug("Control received: %s", msg)
@@ -336,13 +361,14 @@ class Kernel(SingletonConfigurable):
                 if inspect.isawaitable(result):
                     await result
             except Exception:
-                self.log.error("Exception in control handler:", exc_info=True)
+                self.log.error("Exception in control handler:", exc_info=True)  # noqa: G201
 
         sys.stdout.flush()
         sys.stderr.flush()
         self._publish_status("idle", "control")
         # flush to ensure reply is sent
-        self.control_stream.flush(zmq.POLLOUT)
+        if self.control_stream:
+            self.control_stream.flush(zmq.POLLOUT)
 
     def should_handle(self, stream, msg, idents):
         """Check whether a shell-channel message should be handled
@@ -359,7 +385,8 @@ class Kernel(SingletonConfigurable):
 
     async def dispatch_shell(self, msg):
         """dispatch shell requests"""
-
+        if not self.session:
+            return
         # flush control queue before handling shell requests
         await self._flush_control_queue()
 
@@ -367,7 +394,7 @@ class Kernel(SingletonConfigurable):
         try:
             msg = self.session.deserialize(msg, content=True, copy=False)
         except Exception:
-            self.log.error("Invalid Message", exc_info=True)
+            self.log.error("Invalid Message", exc_info=True)  # noqa: G201
             return
 
         # Set the parent message for side effects.
@@ -382,7 +409,8 @@ class Kernel(SingletonConfigurable):
             self._publish_status("idle", "shell")
             # flush to ensure reply is sent before
             # handling the next request
-            self.shell_stream.flush(zmq.POLLOUT)
+            if self.shell_stream:
+                self.shell_stream.flush(zmq.POLLOUT)
             return
 
         # Print some info about this message and leave a '--->' marker, so it's
@@ -408,7 +436,7 @@ class Kernel(SingletonConfigurable):
                 if inspect.isawaitable(result):
                     await result
             except Exception:
-                self.log.error("Exception in message handler:", exc_info=True)
+                self.log.error("Exception in message handler:", exc_info=True)  # noqa: G201
             except KeyboardInterrupt:
                 # Ctrl-c shouldn't crash the kernel here.
                 self.log.error("KeyboardInterrupt caught in kernel.")
@@ -423,7 +451,8 @@ class Kernel(SingletonConfigurable):
         self._publish_status("idle", "shell")
         # flush to ensure reply is sent before
         # handling the next request
-        self.shell_stream.flush(zmq.POLLOUT)
+        if self.shell_stream:
+            self.shell_stream.flush(zmq.POLLOUT)
 
     def pre_handler_hook(self):
         """Hook to execute before calling message handler"""
@@ -443,7 +472,7 @@ class Kernel(SingletonConfigurable):
             self.log.info("Exiting as there is no eventloop")
             return
 
-        def advance_eventloop():
+        async def advance_eventloop():
             # check if eventloop changed:
             if self.eventloop is not eventloop:
                 self.log.info("exiting eventloop %s", eventloop)
@@ -459,17 +488,19 @@ class Kernel(SingletonConfigurable):
             except KeyboardInterrupt:
                 # Ctrl-C shouldn't crash the kernel
                 self.log.error("KeyboardInterrupt caught in kernel")
-                pass
             if self.eventloop is eventloop:
                 # schedule advance again
                 schedule_next()
 
         def schedule_next():
             """Schedule the next advance of the eventloop"""
-            # flush the eventloop every so often,
-            # giving us a chance to handle messages in the meantime
+            # call_later allows the io_loop to process other events if needed.
+            # Going through schedule_dispatch ensures all other dispatches on msg_queue
+            # are processed before we enter the eventloop, even if the previous dispatch was
+            # already consumed from the queue by process_one and the queue is
+            # technically empty.
             self.log.debug("Scheduling eventloop advance")
-            self.io_loop.call_later(0.001, advance_eventloop)
+            self.io_loop.call_later(0.001, partial(self.schedule_dispatch, advance_eventloop))
 
         # begin polling the eventloop
         schedule_next()
@@ -483,7 +514,8 @@ class Kernel(SingletonConfigurable):
             This is now a coroutine
         """
         # flush messages off of shell stream into the message queue
-        self.shell_stream.flush()
+        if self.shell_stream:
+            self.shell_stream.flush()
         # process at most one shell message per iteration
         await self.process_one(wait=False)
 
@@ -498,7 +530,7 @@ class Kernel(SingletonConfigurable):
             try:
                 t, dispatch, args = self.msg_queue.get_nowait()
             except (asyncio.QueueEmpty, QueueEmpty):
-                return None
+                return
         await dispatch(*args)
 
     async def dispatch_queue(self):
@@ -543,19 +575,20 @@ class Kernel(SingletonConfigurable):
         self.msg_queue: Queue[t.Any] = Queue()
         self.io_loop.add_callback(self.dispatch_queue)
 
-        self.control_stream.on_recv(self.dispatch_control, copy=False)
+        if self.control_stream:
+            self.control_stream.on_recv(self.dispatch_control, copy=False)
 
         control_loop = self.control_thread.io_loop if self.control_thread else self.io_loop
 
         asyncio.run_coroutine_threadsafe(self.poll_control_queue(), control_loop.asyncio_loop)
-
-        self.shell_stream.on_recv(
-            partial(
-                self.schedule_dispatch,
-                self.dispatch_shell,
-            ),
-            copy=False,
-        )
+        if self.shell_stream:
+            self.shell_stream.on_recv(
+                partial(
+                    self.schedule_dispatch,
+                    self.dispatch_shell,
+                ),
+                copy=False,
+            )
 
         # publish idle status
         self._publish_status("starting", "shell")
@@ -574,7 +607,8 @@ class Kernel(SingletonConfigurable):
 
     def _publish_execute_input(self, code, parent, execution_count):
         """Publish the code request on the iopub stream."""
-
+        if not self.session:
+            return
         self.session.send(
             self.iopub_socket,
             "execute_input",
@@ -585,6 +619,8 @@ class Kernel(SingletonConfigurable):
 
     def _publish_status(self, status, channel, parent=None):
         """send status (busy/idle) on IOPub"""
+        if not self.session:
+            return
         self.session.send(
             self.iopub_socket,
             "status",
@@ -594,11 +630,13 @@ class Kernel(SingletonConfigurable):
         )
 
     def _publish_debug_event(self, event):
+        if not self.session:
+            return
         self.session.send(
             self.iopub_socket,
             "debug_event",
             event,
-            parent=self.get_parent("control"),
+            parent=self.get_parent(),
             ident=self._topic("debug_event"),
         )
 
@@ -614,7 +652,7 @@ class Kernel(SingletonConfigurable):
         self._parent_ident[channel] = ident
         self._parents[channel] = parent
 
-    def get_parent(self, channel="shell"):
+    def get_parent(self, channel=None):
         """Get the parent request associated with a channel.
 
         .. versionadded:: 6
@@ -629,6 +667,14 @@ class Kernel(SingletonConfigurable):
         message : dict
             the parent message for the most recent request on the channel.
         """
+
+        if channel is None:
+            # If a channel is not specified, get information from current thread
+            if threading.current_thread().name == CONTROL_THREAD_NAME:
+                channel = "control"
+            else:
+                channel = "shell"
+
         return self._parents.get(channel, {})
 
     def send_response(
@@ -641,7 +687,7 @@ class Kernel(SingletonConfigurable):
         track=False,
         header=None,
         metadata=None,
-        channel="shell",
+        channel=None,
     ):
         """Send a response to the message we're currently processing.
 
@@ -651,6 +697,8 @@ class Kernel(SingletonConfigurable):
         This relies on :meth:`set_parent` having been called for the current
         message.
         """
+        if not self.session:
+            return None
         return self.session.send(
             stream,
             msg_or_type,
@@ -683,13 +731,17 @@ class Kernel(SingletonConfigurable):
 
     async def execute_request(self, stream, ident, parent):
         """handle an execute_request"""
+        if not self.session:
+            return
         try:
             content = parent["content"]
             code = content["code"]
-            silent = content["silent"]
+            silent = content.get("silent", False)
             store_history = content.get("store_history", not silent)
             user_expressions = content.get("user_expressions", {})
             allow_stdin = content.get("allow_stdin", False)
+            cell_meta = parent.get("metadata", {})
+            cell_id = cell_meta.get("cellId")
         except Exception:
             self.log.error("Got bad msg: ")
             self.log.error("%s", parent)
@@ -705,25 +757,22 @@ class Kernel(SingletonConfigurable):
             self.execution_count += 1
             self._publish_execute_input(code, parent, self.execution_count)
 
-        cell_id = (parent.get("metadata") or {}).get("cellId")
+        # Arguments based on the do_execute signature
+        do_execute_args = {
+            "code": code,
+            "silent": silent,
+            "store_history": store_history,
+            "user_expressions": user_expressions,
+            "allow_stdin": allow_stdin,
+        }
 
-        if _accepts_cell_id(self.do_execute):
-            reply_content = self.do_execute(
-                code,
-                silent,
-                store_history,
-                user_expressions,
-                allow_stdin,
-                cell_id=cell_id,
-            )
-        else:
-            reply_content = self.do_execute(
-                code,
-                silent,
-                store_history,
-                user_expressions,
-                allow_stdin,
-            )
+        if self._do_exec_accepted_params["cell_meta"]:
+            do_execute_args["cell_meta"] = cell_meta
+        if self._do_exec_accepted_params["cell_id"]:
+            do_execute_args["cell_id"] = cell_id
+
+        # Call do_execute with the appropriate arguments
+        reply_content = self.do_execute(**do_execute_args)
 
         if inspect.isawaitable(reply_content):
             reply_content = await reply_content
@@ -741,7 +790,7 @@ class Kernel(SingletonConfigurable):
         reply_content = json_clean(reply_content)
         metadata = self.finish_metadata(parent, metadata, reply_content)
 
-        reply_msg = self.session.send(
+        reply_msg: dict[str, t.Any] = self.session.send(  # type:ignore[assignment]
             stream,
             "execute_reply",
             reply_content,
@@ -763,6 +812,7 @@ class Kernel(SingletonConfigurable):
         user_expressions=None,
         allow_stdin=False,
         *,
+        cell_meta=None,
         cell_id=None,
     ):
         """Execute user code. Must be overridden by subclasses."""
@@ -770,6 +820,8 @@ class Kernel(SingletonConfigurable):
 
     async def complete_request(self, stream, ident, parent):
         """Handle a completion request."""
+        if not self.session:
+            return
         content = parent["content"]
         code = content["code"]
         cursor_pos = content["cursor_pos"]
@@ -793,6 +845,8 @@ class Kernel(SingletonConfigurable):
 
     async def inspect_request(self, stream, ident, parent):
         """Handle an inspect request."""
+        if not self.session:
+            return
         content = parent["content"]
 
         reply_content = self.do_inspect(
@@ -815,6 +869,8 @@ class Kernel(SingletonConfigurable):
 
     async def history_request(self, stream, ident, parent):
         """Handle a history request."""
+        if not self.session:
+            return
         content = parent["content"]
 
         reply_content = self.do_history(**content)
@@ -842,7 +898,9 @@ class Kernel(SingletonConfigurable):
 
     async def connect_request(self, stream, ident, parent):
         """Handle a connect request."""
-        content = self._recorded_ports.copy() if self._recorded_ports is not None else {}
+        if not self.session:
+            return
+        content = self._recorded_ports.copy() if self._recorded_ports else {}
         content["status"] = "ok"
         msg = self.session.send(stream, "connect_reply", content, parent, ident)
         self.log.debug("%s", msg)
@@ -860,6 +918,8 @@ class Kernel(SingletonConfigurable):
 
     async def kernel_info_request(self, stream, ident, parent):
         """Handle a kernel info request."""
+        if not self.session:
+            return
         content = {"status": "ok"}
         content.update(self.kernel_info)
         msg = self.session.send(stream, "kernel_info_reply", content, parent, ident)
@@ -867,6 +927,8 @@ class Kernel(SingletonConfigurable):
 
     async def comm_info_request(self, stream, ident, parent):
         """Handle a comm info request."""
+        if not self.session:
+            return
         content = parent["content"]
         target_name = content.get("target_name", None)
 
@@ -883,7 +945,7 @@ class Kernel(SingletonConfigurable):
         msg = self.session.send(stream, "comm_info_reply", reply_content, parent, ident)
         self.log.debug("%s", msg)
 
-    def _send_interupt_children(self):
+    def _send_interrupt_children(self):
         if os.name == "nt":
             self.log.error("Interrupt message not supported on Windows")
         else:
@@ -894,23 +956,36 @@ class Kernel(SingletonConfigurable):
             if pgid and pgid == pid and hasattr(os, "killpg"):
                 try:
                     os.killpg(pgid, SIGINT)
-                    return
                 except OSError:
-                    pass
-            try:
+                    os.kill(pid, SIGINT)
+                    raise
+            else:
                 os.kill(pid, SIGINT)
-            except OSError:
-                pass
 
     async def interrupt_request(self, stream, ident, parent):
         """Handle an interrupt request."""
-        self._send_interupt_children()
-        content = parent["content"]
+        if not self.session:
+            return
+        content: dict[str, t.Any] = {"status": "ok"}
+        try:
+            self._send_interrupt_children()
+        except OSError as err:
+            import traceback
+
+            content = {
+                "status": "error",
+                "traceback": traceback.format_stack(),
+                "ename": str(type(err).__name__),
+                "evalue": str(err),
+            }
+
         self.session.send(stream, "interrupt_reply", content, parent, ident=ident)
         return
 
     async def shutdown_request(self, stream, ident, parent):
         """Handle a shutdown request."""
+        if not self.session:
+            return
         content = self.do_shutdown(parent["content"]["restart"])
         if inspect.isawaitable(content):
             content = await content
@@ -921,12 +996,14 @@ class Kernel(SingletonConfigurable):
         await self._at_shutdown()
 
         self.log.debug("Stopping control ioloop")
-        control_io_loop = self.control_stream.io_loop
-        control_io_loop.add_callback(control_io_loop.stop)
+        if self.control_stream:
+            control_io_loop = self.control_stream.io_loop
+            control_io_loop.add_callback(control_io_loop.stop)
 
         self.log.debug("Stopping shell ioloop")
-        shell_io_loop = self.shell_stream.io_loop
-        shell_io_loop.add_callback(shell_io_loop.stop)
+        if self.shell_stream:
+            shell_io_loop = self.shell_stream.io_loop
+            shell_io_loop.add_callback(shell_io_loop.stop)
 
     def do_shutdown(self, restart):
         """Override in subclasses to do things when the frontend shuts down the
@@ -936,6 +1013,8 @@ class Kernel(SingletonConfigurable):
 
     async def is_complete_request(self, stream, ident, parent):
         """Handle an is_complete request."""
+        if not self.session:
+            return
         content = parent["content"]
         code = content["code"]
 
@@ -952,6 +1031,8 @@ class Kernel(SingletonConfigurable):
 
     async def debug_request(self, stream, ident, parent):
         """Handle a debug request."""
+        if not self.session:
+            return
         content = parent["content"]
         reply_content = self.do_debug_request(content)
         if inspect.isawaitable(reply_content):
@@ -966,22 +1047,25 @@ class Kernel(SingletonConfigurable):
             metric_value = getattr(process, name)()
             if attribute is not None:  # ... a named tuple
                 return getattr(metric_value, attribute)
-            else:  # ... or a number
-                return metric_value
+            # ... or a number
+            return metric_value
         # Avoid littering logs with stack traces
         # complaining about dead processes
         except BaseException:
-            return None
+            return 0
 
     async def usage_request(self, stream, ident, parent):
         """Handle a usage request."""
+        if not self.session:
+            return
         reply_content = {"hostname": socket.gethostname(), "pid": os.getpid()}
         current_process = psutil.Process()
         all_processes = [current_process, *current_process.children(recursive=True)]
         # Ensure 1) self.processes is updated to only current subprocesses
         # and 2) we reuse processes when possible (needed for accurate CPU)
         self.processes = {
-            process.pid: self.processes.get(process.pid, process) for process in all_processes
+            process.pid: self.processes.get(process.pid, process)  # type:ignore[misc,call-overload]
+            for process in all_processes
         }
         reply_content["kernel_cpu"] = sum(
             [
@@ -999,7 +1083,7 @@ class Kernel(SingletonConfigurable):
         cpu_percent = psutil.cpu_percent()
         # https://psutil.readthedocs.io/en/latest/index.html?highlight=cpu#psutil.cpu_percent
         # The first time cpu_percent is called it will return a meaningless 0.0 value which you are supposed to ignore.
-        if cpu_percent is not None and cpu_percent != 0.0:
+        if cpu_percent is not None and cpu_percent != 0.0:  # type:ignore[redundant-expr]
             reply_content["host_cpu_percent"] = cpu_percent
         reply_content["cpu_count"] = psutil.cpu_count(logical=True)
         reply_content["host_virtual_memory"] = dict(psutil.virtual_memory()._asdict())
@@ -1021,7 +1105,7 @@ class Kernel(SingletonConfigurable):
             bufs = parent["buffers"]
             msg_id = parent["header"]["msg_id"]
         except Exception:
-            self.log.error("Got bad msg: %s", parent, exc_info=True)
+            self.log.error("Got bad msg: %s", parent, exc_info=True)  # noqa: G201
             return
 
         md = self.init_metadata(parent)
@@ -1033,7 +1117,8 @@ class Kernel(SingletonConfigurable):
         sys.stderr.flush()
 
         md = self.finish_metadata(parent, md, reply_content)
-
+        if not self.session:
+            return
         self.session.send(
             stream,
             "apply_reply",
@@ -1066,6 +1151,8 @@ class Kernel(SingletonConfigurable):
             self.aborted.add(str(mid))
 
         content = dict(status="ok")
+        if not self.session:
+            return
         reply_msg = self.session.send(
             stream, "abort_reply", content=content, parent=parent, ident=ident
         )
@@ -1077,7 +1164,8 @@ class Kernel(SingletonConfigurable):
             "clear_request is deprecated in kernel_base. It is only part of IPython parallel"
         )
         content = self.do_clear()
-        self.session.send(stream, "clear_reply", ident=idents, parent=parent, content=content)
+        if self.session:
+            self.session.send(stream, "clear_reply", ident=idents, parent=parent, content=content)
 
     def do_clear(self):
         """DEPRECATED since 4.0.3"""
@@ -1103,7 +1191,8 @@ class Kernel(SingletonConfigurable):
 
         # flush streams, so all currently waiting messages
         # are added to the queue
-        self.shell_stream.flush()
+        if self.shell_stream:
+            self.shell_stream.flush()
 
         # Callback to signal that we are done aborting
         # dispatch functions _must_ be async
@@ -1116,13 +1205,24 @@ class Kernel(SingletonConfigurable):
         # before we reset the flag
         schedule_stop_aborting = partial(self.schedule_dispatch, stop_aborting)
 
-        # if we have a delay, give messages this long to arrive on the queue
-        # before we stop aborting requests
-        asyncio.get_event_loop().call_later(self.stop_on_error_timeout, schedule_stop_aborting)
+        if self.stop_on_error_timeout:
+            # if we have a delay, give messages this long to arrive on the queue
+            # before we stop aborting requests
+            self.io_loop.call_later(self.stop_on_error_timeout, schedule_stop_aborting)
+            # If we have an eventloop, it may interfere with the call_later above.
+            # If the loop has a _schedule_exit method, we call that so the loop exits
+            # after stop_on_error_timeout, returning to the main io_loop and letting
+            # the call_later fire.
+            if self.eventloop is not None and hasattr(self.eventloop, "_schedule_exit"):
+                self.eventloop._schedule_exit(self.stop_on_error_timeout + 0.01)
+        else:
+            schedule_stop_aborting()
 
     def _send_abort_reply(self, stream, msg, idents):
         """Send a reply to an aborted request"""
-        self.log.info(f"Aborting {msg['header']['msg_id']}: {msg['header']['msg_type']}")
+        if not self.session:
+            return
+        self.log.info("Aborting %s: %s", msg["header"]["msg_id"], msg["header"]["msg_type"])
         reply_type = msg["header"]["msg_type"].rsplit("_", 1)[0] + "_reply"
         status = {"status": "aborted"}
         md = self.init_metadata(msg)
@@ -1198,10 +1298,10 @@ class Kernel(SingletonConfigurable):
             except zmq.ZMQError as e:
                 if e.errno == zmq.EAGAIN:
                     break
-                else:
-                    raise
+                raise
 
         # Send the input request.
+        assert self.session is not None
         content = json_clean(dict(prompt=prompt, password=password))
         self.session.send(self.stdin_socket, "input_request", content, parent, ident=ident)
 
@@ -1227,7 +1327,7 @@ class Kernel(SingletonConfigurable):
                 self.log.warning("Invalid Message:", exc_info=True)
 
         try:
-            value = reply["content"]["value"]
+            value = reply["content"]["value"]  # type:ignore[index]
         except Exception:
             self.log.error("Bad input_reply: %s", parent)
             value = ""
@@ -1243,8 +1343,9 @@ class Kernel(SingletonConfigurable):
         Like `killpg`, but does not include the current process
         (or possible parents).
         """
+        sig_rep = f"{Signals(signum)!r}"
         for p in self._process_children():
-            self.log.debug(f"Sending {Signals(signum)!r} to subprocess {p}")
+            self.log.debug("Sending %s to subprocess %s", sig_rep, p)
             try:
                 if signum == SIGTERM:
                     p.terminate()
@@ -1293,7 +1394,9 @@ class Kernel(SingletonConfigurable):
                 # signals only children, not current process
                 self._signal_children(signum)
                 self.log.debug(
-                    f"Will sleep {delay}s before checking for children and retrying. {children}"
+                    "Will sleep %s sec before checking for children and retrying. %s",
+                    delay,
+                    children,
                 )
                 await asyncio.sleep(delay)
 
@@ -1305,11 +1408,12 @@ class Kernel(SingletonConfigurable):
             self.log.exception("Exception during subprocesses termination %s", e)
 
         finally:
-            if self._shutdown_message is not None:
+            if self._shutdown_message is not None and self.session:
                 self.session.send(
                     self.iopub_socket,
                     self._shutdown_message,
                     ident=self._topic("shutdown"),
                 )
                 self.log.debug("%s", self._shutdown_message)
-            self.control_stream.flush(zmq.POLLOUT)
+            if self.control_stream:
+                self.control_stream.flush(zmq.POLLOUT)
